import java.io.*;
import java.util.ArrayList;
import java.util.PriorityQueue;
import java.util.TreeMap;

/**
 * Compress and decompress text files
 * @author Logan Tucker, Winter 20222, cs10 code
 * @author Tucker Simpson, Winter 20222, cs10 code
 */

public class Huff {
    public String pathName;

    BinaryTree<CnF> codeTreeRoot;


    public Huff(String pathName) {
        this.pathName = pathName;
    }

    public static void main(String[] args) throws Exception {
        Huff test = new Huff("cs10/inputs/USConstitution.txt");
        System.out.println("File found");
        test.compress();
        System.out.println("Compression done");
        test.decompress();
        System.out.println("Decompression done");
    }

    /**
     * Return list of all characters in file
     */
    public ArrayList<Character> getChrList() {
        ArrayList<Character> chrList = new ArrayList<>();
        BufferedReader input;
        try {
            // Contain file-reading in try-catch
            input = new BufferedReader(new FileReader(pathName));

            // Add characters to arrayList
            int c = input.read();
            while (c != -1) {
                chrList.add((char) c);
                c = input.read();
            }
            input.close();
        } catch(Exception FileNotFoundException){
            System.out.println("File "+pathName+" not found");
            return null;
        }

        return chrList;

    }

    /**
     * @return frequency table of characters in chrList
     */
    public TreeMap<Character, Integer> getFrqMap() {

        TreeMap<Character, Integer> frqMap = new TreeMap<>();
        ArrayList<Character> chrList = getChrList();
        if (chrList == null) return null;

        // Loop through characters in the list, adding to a Map with value equal to number of occurrences in file
        for (Character c : chrList) {
            if (frqMap.containsKey(c)) {
                frqMap.put(c, frqMap.get(c) + 1);
            }
            else {
                frqMap.put(c, 1);
            }
        }
        return frqMap;
    }

    /**
     *
     * @return return priority queue of characters, sorted according to frequency
     */
    public PriorityQueue<BinaryTree<CnF>> getLeafQueue() {
        TreeMap<Character, Integer> frqMap = getFrqMap();
        if (frqMap == null) return null;

        PriorityQueue<BinaryTree<CnF>> tqueue = new PriorityQueue<>(new CnFComparator());

        // Put Chars and Freqs into the Priority que, which sorts according to frequency of characters
        for (Character c : frqMap.keySet()) {
            CnF newCnF = new CnF(c, frqMap.get(c));
            BinaryTree<CnF> newLeaf = new BinaryTree<>(newCnF);

            tqueue.add(newLeaf);
        }
        return tqueue;
    }

    /**
     * @return tree with Huffman encoding based on priority queue of characters
     */
    public BinaryTree<CnF> getCodeTree() {
        PriorityQueue<BinaryTree<CnF>> tqueue = getLeafQueue();
        if(getLeafQueue() == null) return null;

        // Edge case in which input file only has one character (either repeated or occurring once)
        if(tqueue.size()==1){
            BinaryTree<CnF> node = tqueue.remove();
            codeTreeRoot = new BinaryTree<>(new CnF(node.data.frq), null, node);
            return codeTreeRoot;
        }

        // Otherwise, loop through the priority que until it is empty
        while (tqueue.size() > 1) {
            // Extract lowest-frequency trees from queue
            BinaryTree<CnF> low1 = tqueue.remove();
            BinaryTree<CnF> low2 = tqueue.remove();

            // Create a new tree with low1 and low2 as subtrees, and value that equals freq of low1 + freq of low2
            BinaryTree<CnF> newRoot = new BinaryTree<>(new CnF(low1.data.frq + low2.data.frq), low1, low2);

            // Add new tree back to the priority queue
            tqueue.add(newRoot);
        }

        // Edge case accounting for empty input file
        if(tqueue.isEmpty()) {
            codeTreeRoot = null;
        }
        else {
            codeTreeRoot = tqueue.remove();
        }

        // return Huffin code tree
        return codeTreeRoot;
    }

    /**
     * Helper function for GetWordCodes
     *
     * @param codes map holding codes for every character
     * @param node current node we are
     * @param pathSoFar string holding path of 0's and 1's
     * @return map of codes for each character
     */
    public TreeMap<Character, String> codeMiner(TreeMap<Character, String> codes, BinaryTree<CnF> node, String pathSoFar) {
        if(node == null) return null;
        // add path that it took to get to the leaf to TreeMap
        if (node.isLeaf()) {
            codes.put(node.data.chr, pathSoFar);
        }
        else {
            if (node.hasLeft()) {
                // recursive call, continue hunting for leafs
                codeMiner(codes, node.getLeft(), pathSoFar + "0");
            }
            if (node.hasRight()) {
                // recursive call, continue hunting for leafs
                codeMiner(codes, node.getRight(), pathSoFar + "1");
            }
        }
        return codes;
    }

    /**
     * calls getCodeTree and codeMiner
     * @returns TreeMap of codes for each character
     */
    public TreeMap<Character, String> getWordCodes() {
        BinaryTree<CnF> node = getCodeTree();
        TreeMap<Character, String> codes = new TreeMap<>();
        String pathSoFar = "";

        return codeMiner(codes, node, pathSoFar);

    }

    /**
     * Compresses input file and writes into new file
     * @throws IOException
     */
    public void compress() throws IOException {
        BufferedReader input = new BufferedReader(new FileReader(pathName));
        BufferedBitWriter bitOutput = new BufferedBitWriter("cs10/outputs/" + pathName.substring(11, pathName.length() - 4) + "_compressed.txt");
        TreeMap<Character, String> codes = getWordCodes();  // Find codes of each character

        int c = input.read();
        while (c != -1) {
            String code = codes.get((char) c);      // Get code of character

            // Write sequence of 0's and 1's into output file
            for(int i = 0; i < code.length(); i++) {
                if (code.charAt(i) == '1') {
                    bitOutput.writeBit(true);
                }
                else {
                    bitOutput.writeBit(false);
                }
            }
            c = input.read();
        }
        bitOutput.close();
        input.close();
    }


    public void decompress() throws IOException {
        BufferedBitReader bitInput = new BufferedBitReader("cs10/outputs/" + pathName.substring(11, pathName.length() - 4) + "_compressed.txt");
        BufferedWriter output = new BufferedWriter(new FileWriter("cs10/outputs/" + pathName.substring(11, pathName.length() - 4) + "_decompressed.txt"));

        if (bitInput != null){
            BinaryTree<CnF> currNode = codeTreeRoot;

            /*use bits of compress file to navigate code tree, inputting characters into the output file whenever a
                leaf is found*/
            while (bitInput.hasNext()) {

                boolean bit = bitInput.readBit();

                if (bit){
                    currNode = currNode.getRight();
                }
                else {
                    currNode = currNode.getLeft();
                }

                if (currNode.isLeaf()) {
                    output.write(currNode.data.chr);
                    currNode = codeTreeRoot;
                }
            }
        }
        bitInput.close();
        output.close();
    }
}
